#!/project/tcl/bin/wish -f
#
# canvas_ui - a draw-like program
#
set lib [file dirname [info script]]
source $lib/alphabrowser.tk
source $lib/printer.tk
source $lib/size.tk
source $lib/utils.tk
source $lib/colors.tk

alphaBrowser canvas_ui

# So the user can change window size
wm minsize . 64 64

set canvas_ui.news {
The Left button draws objects of
different types, or erases things
when in erase mode.

The Middle button scrolls the canvas.
Shift-Middle scrolls faster.

The Right button moves objects.  Click on an
item to move just it, or drag out a region to
move those items in the region.

Shift-Left tells you whats under the mouse.
Shift-Right does too, but with a wider range.

Control-Left deletes strokes.
}
proc simpleNewsHelper {name} {
    upvar #0 $name.news help
    return $help
}

# Canvas and scrollbars
frame .f -background $backgroundColor
pack append . .f {top expand fill}

set canHeight 400
set canWidth 800
set can [canvas .f.can -background $backgroundColor -width 400 -height 200 \
	    -scrollregion "0 0 $canWidth $canHeight" \
	    -xscrollcommand ".f.xscroll set" \
	    -yscrollcommand ".f.yscroll set" ]

scrollbar .f.xscroll -command "$can xview" -orient horizontal \
    -troughcolor $backgroundColor \
    -background $passiveColor \
    -activebackground $activeColor

scrollbar .f.yscroll -command "$can yview" -orient vertical \
    -troughcolor $backgroundColor \
    -background $passiveColor \
    -activebackground $activeColor

pack append .f .f.xscroll {bottom fillx} .f.yscroll {right filly} \
	.f.can {left expand fill}

# Mode field
labeledEntry .feedback .mode "Mode: " 10 {right}
proc showMode { mode } {
    .feedback.mode.entry delete 0 end
    .feedback.mode.entry insert 0 $mode
}
########################################################################

proc canvasBindings { can } {
# Keybindings to make strokes (Left button)
    bind $can <Button-1> {moveTo %W %x %y}
    bind $can <B1-Motion> {drawTo %W %x %y}
    bind $can <ButtonRelease-1> {drawEnd %W %x %y}

# Keybindings to delete strokes (control-Left)
    bind $can <Control-Button-1> {delete %W %x %y}

# Keybindings to scroll (Middle button, fast or slow drag via Shift)
    bind $can <Button-2> {canvasMark %W %x %y }
    bind $can <B2-Motion> {canvasDrag %W %x %y slow }
    bind $can <Shift-Button-2> {canvasMark %W %x %y }
    bind $can <Shift-B2-Motion> {canvasDrag %W %x %y fast }

# Keybindings to query the canvas (Shift, left or right to determine range)
    bind $can <Shift-Button-1> { feedback [whatsThere %W %x %y 1] }
    bind $can <Shift-Button-3> { feedback [whatsThere %W %x %y 5] }

# Keybindings to move items (Button 3)
    bind $can <Button-3> { itemGrab %W %x %y }
    bind $can <B3-Motion> { itemMove %W %x %y }
    bind $can <ButtonRelease-3> { itemEndGrab %W %x %y }
}

# Modal drawing support

proc setStrokeMode { can } {
    global drawProc endProc mode
    set mode stroke
    set drawProc strokeDraw
    set endProc strokeEnd
    showMode stroke
    $can configure -cursor pencil
}

proc setLineMode { can } {
    global drawProc endProc mode
    set mode stroke
    set drawProc lineDraw
    set endProc lineEnd
    showMode stroke
    $can configure -cursor crosshair
}

proc setBoxMode { can } {
    global drawProc endProc mode
    set mode stroke
    set drawProc boxDraw
    set endProc boxEnd
    showMode stroke
    $can configure -cursor cross
}

proc setOvalMode { can } {
    global drawProc endProc mode
    set mode stroke
    set drawProc ovalDraw
    set endProc ovalEnd
    showMode stroke
    $can configure -cursor target
}

proc setTextMode { can } {
    global drawProc endProc mode
    set mode stroke
    set drawProc textDraw
    set endProc textEnd
    showMode text
    $can configure -cursor tcross
}

proc setEraseMode { can } {
    global drawProc endProc mode
    set mode stroke
    set drawProc eraseDraw
    set endProc eraseEnd
    showMode stroke
    $can configure -cursor dotbox
}

proc setMode {can newMode} {
    # can is ignored, but we should have per-canvas state
    global mode drawProc endProc
    case $newMode {
	stroke {
	    setStrokeMode $can
	}
	line {
	    setLineMode $can
	}
	box {
	    setBoxMode $can
	}
	oval {
	    setOvalMode $can
	}
	text {
	    setTextMode $can
	}
	erase {
	    setEraseMode $can
	}
	default {
	    feedback "Unsupported mode: $newMode"
	    return
	}
    }
    set mode $newMode
    showMode $mode
}

# Generic drawing procedures

proc moveTo { can x y } {
    # Initailize position and segment sequence
    global Segments
    global X Y N
    set X(0) [$can canvasx $x]
    set Y(0) [$can canvasy $y]
    set N 0
    set Segments(0) {}
}

proc drawTo { can x y {proc default} } {
    global drawProc
    global X Y N Segments
    global Move

    if {![info exists N] || $N < 0} {
	return
    }
    if {$proc == "default"} {
	set proc $drawProc
    }
    if {$Move(item) == "SEL"} {
	$can delete $Move(lastBox)
    }
    set x [$can canvasx $x]
    set y [$can canvasy $y]
    set item [$proc $can $x $y]
    set Segments($N) $item
    incr N
    set X($N) $x
    set Y($N) $y
    return $item
}

set itemN 0
proc drawEnd { can x y {proc default} } {
    global endProc itemN N
    if {$proc == "default"} {
	set proc $endProc
    }
    set item [$proc $can $x $y]
    incr itemN
    set N -1
    return $item
}

# Erasing

proc eraseDraw { can x y } {
    global X Y N Segments itemN
    set items [$can find overlapping $X($N) $Y($N) $x $y]
    foreach item $items {
	global canBackground
	if {$item != $canBackground} {
	    $can delete $item
	}
    }
    return {}
}

proc eraseEnd { can x y } {
    drawTo $can $x $y eraseDraw
    return {}
}

# Stroke drawing

proc strokeDraw { can x y } {
    global X Y N
    return [$can create line $X($N) $Y($N) $x $y -tags {line}]
}

proc strokeEnd { can x y } {
    global itemN
    drawTo $can $x $y
    # Replace straight-line segments with bezier curve
    global X Y N Segments
    set createCmd "$can create line"
    for {set i 0} {$i <= $N} {incr i} {
	lappend createCmd $X($i) $Y($i)
	if {$i != $N} {
	    $can delete $Segments($i)
	}
    }
    lappend createCmd -joinstyle round -smooth 1 -tags stroke$itemN
    return [eval $createCmd]
}

# Straight-line drawing

proc lineDraw { can x y } {
    global X Y N Segments itemN
    if {$N > 0} {
	$can delete $Segments([expr $N-1])
    }
    return [$can create line $X(0) $Y(0) $x $y -tags line$itemN]
}

proc lineEnd { can x y } {
    global itemN
    return [drawTo $can $x $y lineDraw]
}

# Box drawing

proc boxDraw { can x y } {
    global X Y N Segments itemN
    if {$N > 0} {
	$can delete $Segments([expr $N-1])
    }
    return [$can create rectangle $X(0) $Y(0) $x $y -tags box$itemN]
}

proc boxEnd { can x y } {
    return [drawTo $can $x $y boxDraw]
}

# Oval drawing

proc ovalDraw { can x y } {
    global X Y N Segments itemN
    if {$N > 0} {
	$can delete $Segments([expr $N-1])
    }
    return [$can create oval $X(0) $Y(0) $x $y -tags oval$itemN]
}

proc ovalEnd { can x y } {
    return [drawTo $can $x $y ovalDraw]
}

# Text drawing

proc textDraw { can x y } {
    global X Y N Segments itemN
    if {$N > 0} {
	$can delete $Segments([expr $N-1])
    }
    return [$can create rectangle $X(0) $Y(0) $x $y -tags textBox$itemN]
}

set tnum 0
proc textEnd { can x y } {
    global Segments X Y N
    drawTo $can $x $y textDraw

    global tnum charWidth charHeight
    incr tnum
    text $can.text$tnum 
    set width [expr {($X($N) > $X(0)) ? ($X($N) - $X(0)) : ($X(0) - $X($N))}]
    set height [expr {($Y($N) > $Y(0)) ? ($Y($N) - $Y(0)) : ($Y(0) - $Y($N))}]
    return [$can create window $X(0) $Y(0) -window $can.text$tnum \
	-width $width -height $height -anchor nw]
}

# Scrolling support for the canvas (direct speed)

proc canvasMark { can x y } {
    global canMarkX
    set canMarkX $x
    $can scan mark $x $y
}
proc canvasDrag { can x y speed } {
    global canMarkX
    if {$speed == "slow"} {
	set x [expr $canMarkX+($x-$canMarkX)/10]
    }
    $can scan dragto $x $y
}

#
# itemGrab selects an item for movement
#
set Move(item) {}
proc itemGrab { can x y } {
    global Move canBackground
    case $Move(item) {
	default {
	    # Normal case, click to select or initiate area select
	    startGrab $can $x $y
	}
	SEL {
	    if {(($Move(X0) < $Move(X1)) && \
		 (($x < $Move(X0)) || ($x > $Move(X1)))) || \
	        (($Move(X0) > $Move(X1)) && \
		 (($x < $Move(X1)) || ($x > $Move(X0)))) || \
	        (($Move(Y0) < $Move(Y1)) && \
		 (($y < $Move(Y0)) || ($y > $Move(Y1)))) || \
	        (($Move(Y0) > $Move(Y1)) && \
		 (($y < $Move(Y1)) || ($y > $Move(Y0))))} {
		$can delete $Move(lastBox)
		startGrab $can $x $y
	    } else {
		set Move(X) $x
		set Move(Y) $y
	    }
	}
    }
}

proc startGrab { can x y } {
    global Move canBackground
    set halo 5
    $can dtag SEL
    set item [$can find closest $x $y $halo]
    if {$item != {} && $item != $canBackground} {
	feedback "Selected item $item [$can gettags $item]"
	set Move(item) $item
	set Move(X) $x
	set Move(Y) $y
    } else {
	feedback "Selecting a region"
	set Move(X0) $x
	set Move(Y0) $y
	set Move(item) DRAG
	moveTo $can $x $y
    }
}

proc itemMove { can x y } {
    global Move
    if {$Move(item) == {}} {
	return
    }
    if {$Move(item) == "DRAG"} {
	# Dragging out a selection
	drawTo $can $x $y boxDraw
	set Move(X1) $x
	set Move(Y1) $y
    } else {
	set dx [expr $x-$Move(X)]
	set dy [expr $y-$Move(Y)]
	$can move $Move(item) $dx $dy
	incr Move(X0) $dx
	incr Move(X1) $dx
	incr Move(Y0) $dy
	incr Move(Y1) $dy
	set Move(X) $x
	set Move(Y) $y
    }
}

proc itemEndGrab { can x y } {
    global Move
    if {![info exists Move(item)] || $Move(item) != "DRAG"} {
	return
    }
    if {$x > $Move(X0)} {
	set Move(X1) $x
    } else {
	set Move(X1) $Move(X0)
	set Move(X0) $x
    }
    if {$y > $Move(Y0)} {
	set Move(Y1) $y
    } else {
	set Move(Y1) $Move(Y0)
	set Move(Y0) $y
    }
    set Move(items) [$can find enclosed $Move(X0) $Move(Y0) $Move(X1) $Move(Y1)]
    set Move(lastBox) [drawEnd $can $x $y boxEnd]
    if {$Move(items) == {}} {
	set Move(item) {}
	$can delete $Move(lastBox)
	feedback "Nothing in selection box"
    } else {
	$can itemconfigure $Move(lastBox) -tags SEL
	$can addtag SEL enclosed $Move(X0) $Move(Y0) $Move(X1) $Move(Y1)
	set Move(item) SEL
	feedback "Move items $Move(items)"
    }
}

# Debugging procs

proc MM { } {
    PA Move
}
proc Segs { } {
    global N Segments
    for {set i 0} {$i < $N} {incr i} {
	puts stdout [format "%s %s" $i $Segments($i)]
    }
}
proc PA { array } {
    global $array
    puts stdout $array
    foreach name [lsort [array names $array]] {
	set ref [format "%s(%s)" $array $name]
	set item [eval "set $ref"]
	puts stdout [format "%8s %s" $name $item]
    }
}
#
# whatsThere displays tags for objects under a mouse click
#
proc whatsThere { can x y halo } {
    set cx [$can canvasx $x]
    set cy [$can canvasy $y]

    set things [$can find overlapping [expr $cx-$halo] [expr $cy-$halo] \
				      [expr $cx+$halo] [expr $cy+$halo]]

    set lasttag {}
    set answer "${things}:"
    foreach item $things {
	set tags [$can gettags $item]
	foreach tag $tags {
	    if {$lasttag == $tag} {
		continue
	    }
	    if {($tag != "Scale") && ($tag != "current")} {
		append answer "$tag "
	    }
	    set lasttag $tag
	}
    }
    return $answer
}
#
# delete things under a mouse click
#
proc delete { can x y {halo 2} } {
    global canBackground

    set cx [$can canvasx $x]
    set cy [$can canvasy $y]

    set item [$can find closest $cx $cy $halo]

    if {($item != {}) && ($item != $canBackground)} {
	$can delete $item
    }
}

#
# outline the canvas scroll region
#
proc showScrollRegion { can } {
    global canWidth canHeight canBackground
    if [info exists canBackground] {
	$can delete $canBackground
    } 
    set canBackground [$can create rectangle 0 0 $canWidth $canHeight \
			-fill white -outline {}]
    $can lower $canBackground
}

proc showScrollRegionOld { can } {
    global canWidth canHeight canBackground
    set items [$can find withtag scrollBorder]
    foreach item $items {
	$can delete $item
    }
    set x1 3 ; set y1 3
    set x2 [expr $canWidth-3]
    set y2 [expr $canHeight-3]
    $can create line $x1 $y1 $x2 $y1 -stipple gray50 -tags {scrollBorder} 
    $can create line $x2 $y1 $x2 $y2 -stipple gray50 -tags {scrollBorder} 
    $can create line $x2 $y2 $x1 $y2 -stipple gray50 -tags {scrollBorder} 
    $can create line $x1 $y2 $x1 $y1 -stipple gray50 -tags {scrollBorder} 
}

canvasBindings $can
setStrokeMode $can

packedButton .buttons .print "Print" "printDialog $can"
packedButton .buttons .size "Size" "sizeDialog $can"

packedButton .buttons .erase "Erase" "setMode $can erase" right
packedButton .buttons .stroke "Stroke" "setMode $can stroke" right
packedButton .buttons .line "Line" "setMode $can line" right
packedButton .buttons .box "Box" "setMode $can box" right
packedButton .buttons .oval "Oval" "setMode $can oval" right
packedButton .buttons .text "Text" "setMode $can text" right

showScrollRegion $can

